# Copyright 2018-present Lenovo
# Confidential and Proprietary

import fileinput
import os
import re
import shlex
import socket


def get_env(scheduler):
    environ = os.environ
    hostname_format = environ.get("HOSTNAME_TRANS_FORMAT", "{node}")
    env = {'pwd': environ['PWD']}
    if scheduler == 'slurm':
        env['nodelist'] = get_host_list(
            str(environ["SLURM_JOB_NODELIST"]), scheduler)
        env['job_nodelist'] = [
            hostname_format.format(node=node) for node in env['nodelist']
        ]
        env['job_submit_name'] = str(environ["SLURM_JOB_NAME"])
        env['job_id'] = str(environ["SLURM_JOBID"])
        env['log_path'] = str(environ["SLURM_SUBMIT_DIR"])
        if "SLURM_NTASKS" not in environ:
            env['ntasks'] = 2
        else:
            env['ntasks'] = int(environ["SLURM_NTASKS"])
        if 'SLURM_CPUS_ON_NODE' in environ:
            env['cpu_curr_node'] = int(environ["SLURM_CPUS_ON_NODE"])
        if 'SLURM_JOB_CPUS_PER_NODE' in environ:
            find_min_max_cpu(environ["SLURM_JOB_CPUS_PER_NODE"], env)
        if 'SLURM_CPUS_PER_TASK' in environ:
            env['cpu_per_task'] = int(environ["SLURM_CPUS_PER_TASK"])
    elif scheduler == 'pbs':
        env['nodelist'] = get_host_list(
            fileinput.input(environ['PBS_NODEFILE']), scheduler)
        env['job_nodelist'] = [
            hostname_format.format(node=node) for node in env['nodelist']
        ]
        env['job_submit_name'] = str(environ["PBS_JOBNAME"])
        env['job_id'] = str(environ["PBS_JOBID"])
        env['log_path'] = str(environ["PBS_O_WORKDIR"])
        env['ntasks'] = len(env['nodelist'])
        env['cpu_per_task'] = int(environ["NCPUS"])
    else:
        env['nodelist'] = get_host_list(environ["LSB_MCPU_HOSTS"], scheduler)
        env['job_nodelist'] = [
            hostname_format.format(node=node) for node in env['nodelist']
        ]
        env['job_submit_name'] = str(environ["LSB_JOBNAME"])
        env['job_id'] = str(environ["LSB_JOBID"])
        env['log_path'] = str(environ["LSB_OUTDIR"])
        env['ntasks'] = len(env['job_nodelist'])
        env['cpu_per_task'] = environ["LSB_MCPU_HOSTS"].split(' ')[1]

    cuda_devs = environ.get("CUDA_VISIBLE_DEVICES")
    if not cuda_devs or cuda_devs == 'NoDevFiles':
        env['gpu_per_node'] = 0
    else:
        env['gpu_per_node'] = len(cuda_devs.split(','))

    return env


def find_min_max_cpu(slurm_job_cpus_per_node, env):
    cpu_list = slurm_job_cpus_per_node.split(",")
    cpu_num_list = []
    match_re = re.compile('^(?P<cpu_num>[0-9]+).*')
    for cpu in cpu_list:
        out = match_re.match(cpu)
        if not out:
            return
        cpu_num = int(out.groupdict()['cpu_num'])
        cpu_num_list.append(cpu_num)
    if len(cpu_num_list):
        env['min_cpu_on_node'] = min(cpu_num_list)
        env['max_cpu_on_node'] = max(cpu_num_list)


# singularity exec --nv -B
# /home/hpcadmin /home/lico/ai/container/tensorflow-gpu.image
# python

def get_image_command(env, program, workspace, image, gpu_type):
    ext = os.path.splitext(program)[-1]
    singularity_cmd = ""
    if image != "":
        if env["gpu_per_node"]:
            if gpu_type == "intel":
                singularity_cmd = 'singularity exec {0} --pwd {1} {2}'.format(
                    workspace, env['pwd'], image
                )
            else:
                singularity_cmd = 'singularity exec --nv {0} --pwd \
                          {1} {2}'.format(workspace, env['pwd'], image)

        else:
            singularity_cmd = 'singularity exec {0} --pwd {1} {2}'.format(
                workspace, env['pwd'], image
            )

    commands_dir = {
        '.py': '{0} python'.format(singularity_cmd),
        '.pyc': '{0} python'.format(singularity_cmd),
        '.sh': '{0} bash'.format(singularity_cmd)
    }
    return commands_dir.get(ext)


def get_scheduler_command(scheduler, dis=False):
    gpu_command = ''
    cpu_command = ''
    if scheduler == 'slurm':
        if dis:
            # gpu_command = ' srun -N1 -n1 --gres=gpu:{1} ' \
            #               '--nodelist={2} -l {3}'
            gpu_command = ' srun -N1 -n1 ' \
                          '--nodelist={2} -l {3}'
            cpu_command = ' srun -N1 -n1 --nodelist={1} -l {2}'
        else:
            gpu_command = ' srun -N1 -n1 --cpus-per-task={0} ' \
                          '--cpu_bind=cores --gres=gpu:{1} --nodelist={2} -l {3}'
            cpu_command = ' srun -N1 -n1 --cpus-per-task={0} ' \
                          '--cpu_bind=cores --nodelist={1} -l {2}'
    elif scheduler == 'pbs':
        gpu_command = ' pbs_tmrsh {2} {3}'
        cpu_command = ' pbs_tmrsh {1} {2}'
    elif scheduler == 'lsf':
        gpu_command = ' lsrun -m {2} {3}'
        cpu_command = ' lsrun -m {1} {2}'
    else:
        pass

    return gpu_command, cpu_command


def get_command(scheduler, cpu, gpu, host, command, dis=False):
    gpu_command, cpu_command = get_scheduler_command(
        scheduler, dis
    )
    if gpu:
        cmd = gpu_command.format(cpu, gpu, host, command)
    else:
        cmd = cpu_command.format(cpu, host, command)
    cmd = shlex.split(cmd)
    return cmd


def get_host_list(hosts_str, scheduler):
    if scheduler == 'slurm':
        hosts = get_slurm_host_list(hosts_str)
    elif scheduler == 'pbs':
        hosts = list(set([line.strip('\n') for line in hosts_str]))
    else:
        hosts = hosts_str.strip().split(' ')[::2]
    return hosts


def get_slurm_host_list(hosts_str):
    init_hosts = []
    bracket_num = 0
    part_host = ""

    for single_char in hosts_str + ",":
        if single_char == "," and bracket_num == 0:
            # Comma at top level, split!
            if part_host:
                init_hosts.extend(expand_slurm(part_host))
            part_host = ""
        else:
            part_host += single_char
        if single_char == "[":
            bracket_num += 1
        elif single_char == "]":
            bracket_num -= 1

        if bracket_num > 1 or bracket_num < 0:
            raise Exception("Parse slurm nodelist error. "
                            "Reason: brackets unmatch")
    if bracket_num > 0:
        raise Exception("Parse slurm nodelist error. "
                        "Reason: brackets unmatch")

    exist_hosts = set()
    hosts = []
    for first_result in init_hosts:
        if first_result not in exist_hosts:
            hosts.append(first_result)
            exist_hosts.add(first_result)
    return hosts


def expand_slurm(need_expand):
    if need_expand == "":
        return [""]
    m = re.match(r'([^,\[]*)(\[[^\]]*\])?(.*)', need_expand)
    (prefix, rangelist, rest) = m.group(1, 2, 3)

    rest_expand = expand_slurm(rest)

    if not rangelist:
        curr_expand = [prefix]
    else:
        curr_expand = []
        for range_slurm in rangelist[1:-1].split(","):
            results = expand_slurm_range(prefix, range_slurm)
            curr_expand.extend(results)
    if len(curr_expand) * len(rest_expand) > 10000:
        raise Exception("Parse slurm nodelist error. "
                        "Reason: nodelist too long")
    return [curr_part + rest_part
            for curr_part in curr_expand
            for rest_part in rest_expand]


def expand_slurm_range(prefix, range_slurm):
    m = re.match(r'^[0-9]+$', range_slurm)
    if m:
        return ["%s%s" % (prefix, range_slurm)]
    match_re = re.match(r'^([0-9]+)-([0-9]+)$', range_slurm)
    if not match_re:
        raise Exception("Parse slurm nodelist error. "
                        "Reason: nodelist range error")
    (exp_low, exp_high) = match_re.group(1, 2)
    low = int(exp_low)
    high = int(exp_high)
    width = len(exp_low)
    if high < low:
        raise Exception("Parse slurm nodelist error. "
                        "Reason: nodelist start larger than stop")
    elif high - low > 10000:
        raise Exception("Parse slurm nodelist error. "
                        "Reason: nodelist range too large")
    results = []
    for i in range(low, high + 1):
        results.append("%s%0*d" % (prefix, width, i))
    return results


def get_hosts_with_ports(node_list, ps_num=1, worker_num=1):
    nodes_list = node_list
    ps_hosts = []
    worker_hosts = []

    for node in nodes_list:
        port_list = get_port_from_host(node, ps_num + worker_num)
        for i in range(ps_num):
            ps_hosts.append(node + ":" + str(port_list[i]))
        for i in range(ps_num, ps_num + worker_num):
            worker_hosts.append(node + ":" + str(port_list[i]))

    return ps_hosts, worker_hosts


def get_ps_tracker(host, env=None):
    port = get_port_from_host(host, 1)
    env['DMLC_ROLE'] = 'scheduler'
    env['DMLC_PS_ROOT_URI'] = socket.gethostbyname(str(host))
    env['DMLC_PS_ROOT_PORT'] = str(port[0])
    return env


def get_port_from_host(ip, port_num):
    port_list = []
    # If there are available ports
    environ = os.environ
    if 'LICO_AVAILABLE_PORTS' in environ and environ['LICO_AVAILABLE_PORTS']:
        available_ports = environ['LICO_AVAILABLE_PORTS'].split(',')
        for idx in range(min(port_num, len(available_ports))):
            port = int(available_ports.pop())
            port_list.append(port)
        environ['LICO_AVAILABLE_PORTS'] = ','.join(available_ports)
        if len(port_list) == port_num:
            return port_list
    # If no enough available port
    port = 20000
    for port in range(port, port + 1000):
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.settimeout(1)
            status = sock.connect_ex((ip, port))
            if status == 0:
                continue
            else:
                port_list.append(port)
                if len(port_list) == port_num:
                    return port_list
        except socket.error:
            continue


def get_only_worker_hosts(hosts, worker_num):
    node_num = len(hosts)
    worker_per_node = int(worker_num / node_num)

    _, worker_hosts = get_hosts_with_ports(
        hosts[:],
        ps_num=0,
        worker_num=worker_per_node
    )

    return worker_hosts, worker_per_node


def get_ps_worker_hosts(hosts, ps_num, worker_num):
    node_num = len(hosts)
    worker_per_node = int(worker_num / node_num)
    ps_per_node = int(ps_num / node_num)
    # calculate the number of node whose ps number
    #  is ps_num_per_node+1
    ps_plus = ps_num % node_num

    ps_hosts, worker_hosts = get_hosts_with_ports(
        hosts[:ps_plus],
        ps_num=ps_per_node + 1,
        worker_num=worker_per_node)
    tmp_ps_hosts, tmp_worker_hosts = get_hosts_with_ports(
        hosts[ps_plus:],
        ps_num=ps_per_node,
        worker_num=worker_per_node)
    ps_hosts.extend(tmp_ps_hosts)
    worker_hosts.extend(tmp_worker_hosts)
    return ps_hosts, worker_hosts, ps_per_node, ps_plus, worker_per_node


def check_args(args, env):
    import sys
    new_args = {
        '--node-num': len(env['nodelist'])
        if 'nodelist' in env else int(args['--node-num'])
    }
    node_num = new_args['--node-num']
    worker_num = int(args['--worker_number'])

    if worker_num % node_num != 0:
        print("The worker number must be divided by node number")
        sys.exit(1)

    worker_num_per_node = int(worker_num / node_num)

    new_args['--cpu-num'] = env['cpu_curr_node'] \
        if 'cpu_curr_node' in env else int(args['--cpu-num'])
    new_args['--gpu-num'] = env['gpu_per_node'] \
        if 'gpu_per_node' in env else int(args['--gpu-num'])
    if new_args['--gpu-num'] % worker_num_per_node != 0:
        print("The gpu number must be divided by worker number for per node")
        sys.exit(1)
    return new_args
