"""
Lico AI Deep Learning

Copyright 2018-present Lenovo
Confidential and Proprietary

Usage: {program} [options] <command> [<args> ...]

Options:
  -h --help     Show this screen.

Available commands:
  {available_commands}

Environment:

See '{program} help <command>' for help on specific commands.
"""

import os
import sys
import time
from subprocess import Popen, check_call
from threading import Thread
from traceback import print_exc

import docopt_subcommands as dsc
import psutil
from bash import CUDA_BASH
from common import (
    check_args, get_command, get_env, get_image_command, get_only_worker_hosts,
    get_ps_worker_hosts,
)

version = "1.0.0"

sc = dsc.Subcommands(
    doc_template=__doc__, program='lico-dl-run')


def run_deep_learning():
    import sys
    sc(sys.argv[1:])


def check_call_gpu(cmd, gpu_num):
    if gpu_num == 0:
        no_cuda_env = dict(os.environ)
        no_cuda_env["CUDA_VISIBLE_DEVICES"] = ""
        check_call(cmd, env=no_cuda_env)
    else:
        check_call(cmd)


def base_tensorflow(args, env):  # noqa: C901
    cpu_curr_node = int(args['--cpu-num'])
    gpu_per_node = int(args['--gpu-num'])
    worker_num = int(args['--worker_number'])
    ps_num = int(args['--ps_number'])
    node_num = args['--node-num']
    worker_per_node = int(worker_num / node_num)
    gpu_per_worker = int(gpu_per_node / worker_per_node)

    img_cmd = get_image_command(
        env, args['<program>'],
        args['<workspace>'], args['<image>'], args['--gpu_type']
    )
    prog_cmd_template_single = img_cmd + " {} {}"
    prog_cmd_template = \
        img_cmd + " {} --" + args['--ps-args'] + "={} --" + \
        args['--wk-args'] + "={} --" + args['--job-name'] + \
        "={} --" + args['--task-index'] + "={} {}"
    only_work_cmd_template = \
        img_cmd + " {} --" + args['--wk-args'] + \
        "={} --" + args['--job-name'] + "={} --" + \
        args['--task-index'] + "={} {}"

    early_stop_file = ''
    le_args = args.get("--letrain_args")
    if le_args:
        try:
            train_dir = ''
            le_args = le_args.replace('\n', ' ').split()
            for i, item in enumerate(le_args):
                if '--train_dir=' in item:
                    train_dir = item.split('=')[1]
                elif '--train_dir' in item:
                    train_dir = le_args[i + 1]

                if train_dir:
                    early_stop_file = os.path.join(train_dir, 'es_flag')
                    if os.path.exists(early_stop_file):
                        os.remove(early_stop_file)
                    break
        except Exception:
            print_exc("Failed to parse train directory.")

    if ps_num == 0 and worker_num == 1:
        prog_cmd = prog_cmd_template_single.format(
            args['<program>'], args['<args>']
        ).replace('\n', ' ')
        cmd = get_command(args['<scheduler>'], cpu_curr_node,
                          gpu_per_node, env['nodelist'][0], prog_cmd)
        print(cmd)
        sys.stdout.flush()
        check_call_gpu(cmd, gpu_per_node)
    elif ps_num == 0 and worker_num > 1:
        worker_hosts, worker_per_node = \
            get_only_worker_hosts(env['job_nodelist'], worker_num)

        workerlist = []
        for index, value in enumerate(worker_hosts):
            job_name = args['--wk-job-name']
            info = (
                args['<program>'], ",".join(worker_hosts),
                job_name, index, args['<args>']
            )
            prog_cmd = only_work_cmd_template.format(*info)
            if int(gpu_per_node) > 1:
                cuda_template = CUDA_BASH.format(
                    start=index % worker_per_node * gpu_per_worker,
                    num=gpu_per_worker,
                    cmd=prog_cmd
                )
                prog_cmd = 'bash -c "{0}"'.format(cuda_template)
            node_index = int(index / worker_per_node)
            cmd = get_command(
                args['<scheduler>'],
                0,
                gpu_per_worker,
                env['nodelist'][node_index],
                prog_cmd,
                dis=True)
            print(cmd)

            current = Server(cmd, gpu_per_worker)
            workerlist.append(current)
            current.start()
        sys.stdout.flush()
        early_stop_monitor = Thread(
            target=monitor_early_stop,
            args=(workerlist, early_stop_file)
        )
        early_stop_monitor.start()
        update_worklist(workerlist)
    else:
        ps_hosts, worker_hosts, ps_per_node, ps_plus, worker_per_node = \
            get_ps_worker_hosts(env['job_nodelist'], ps_num, worker_num)
        index_plus = (ps_per_node + 1) * ps_plus
        pslist = []
        for index, value in enumerate(ps_hosts):
            job_name = args['--ps-job-name']
            info = (args['<program>'], ",".join(ps_hosts),
                    ",".join(worker_hosts), job_name, index, args['<args>'])
            prog_cmd = prog_cmd_template.format(*info).replace('\n', ' ')
            prog_cmd = 'bash -c "export CUDA_VISIBLE_DEVICES=-1 && {0}"'. \
                format(prog_cmd)
            node_index = int(index / (ps_per_node + 1)) \
                if index < index_plus \
                else ps_plus + int((index - index_plus) / ps_per_node)
            cmd = get_command(
                args['<scheduler>'], '0', '0',
                env['nodelist'][node_index], prog_cmd, dis=True)
            print(cmd)

            current = Server(cmd, 0)
            pslist.append(current)
            current.start()

        workerlist = []
        for index, value in enumerate(worker_hosts):
            job_name = args['--wk-job-name']
            info = (args['<program>'], ",".join(ps_hosts),
                    ",".join(worker_hosts), job_name, index, args['<args>'])
            prog_cmd = prog_cmd_template.format(*info).replace('\n', ' ')
            node_index = int(index / worker_per_node)
            # if args['<scheduler>'] != 'slurm' and int(gpu_per_node) > 1:
            if int(gpu_per_node) > 1:
                cuda_template = CUDA_BASH.format(
                    start=index % worker_per_node * gpu_per_worker,
                    num=gpu_per_worker,
                    cmd=prog_cmd
                )
                prog_cmd = 'bash -c "{0}"'.format(cuda_template)

            cmd = get_command(
                args['<scheduler>'],
                0,
                gpu_per_worker,
                env['nodelist'][node_index],
                prog_cmd,
                dis=True)
            print(cmd)

            current = Server(cmd, gpu_per_worker)
            workerlist.append(current)
            current.start()
        sys.stdout.flush()
        early_stop_monitor = Thread(
            target=monitor_early_stop,
            args=(workerlist, early_stop_file)
        )
        early_stop_monitor.start()
        update_worklist(workerlist)
        for index, value in enumerate(pslist):
            try:
                pslist[index].terminate_job()
            except Exception:
                print_exc('cannot kill ps job')

    # clean inner temp files
    if os.path.exists(early_stop_file):
        os.remove(early_stop_file)


def monitor_early_stop(workerlist, early_stop_file):
    if len(workerlist) and early_stop_file:
        while True:
            time.sleep(2)
            if os.path.exists(early_stop_file):
                for index, value in enumerate(workerlist):
                    try:
                        sys.stdout.flush()
                        workerlist[index].terminate_job()
                    except Exception:
                        print_exc('cannot kill worker')
                break

            stopped_workers_num = 0
            for index, value in enumerate(workerlist):
                if not workerlist[index].is_alive():
                    stopped_workers_num += 1

            if stopped_workers_num == len(workerlist):
                break


def update_worklist(workerlist):
    for index, value in enumerate(workerlist):
        workerlist[index].join()


class Server(object):
    def __init__(self, cmd, gpu_num):
        self.cmd = cmd
        self.p = None
        if gpu_num == 0:
            self.env = dict(os.environ)
            self.env["CUDA_VISIBLE_DEVICES"] = ""
        else:
            self.env = None

    def start(self):
        self.shutdown = False
        self.p = Popen(self.cmd, env=self.env)

    def join(self):
        self.p.wait()

    def is_alive(self):
        return self.p.poll() is None

    def terminate_job(self):
        try:
            parent = psutil.Process(self.p.pid)
        except psutil.NoSuchProcess:
            return

        try:
            for child in parent.children(recursive=True):
                child.kill()
        except Exception:
            pass
        parent.kill()


@sc.command('version')
def run_version(precommand, args):
    """
    Usage: {program} version
    """
    print(version)


@sc.command('letrain')
def run_letrain(precommand, args):
    """Usage:
    {program} {command} <scheduler> <workspace> <image> [options]

    Options:
        --gpu_type GPU_TYPE
                GPU vendor type, if not specified, NVIDIA will be used.
                [default: None]
        --ps_number PS_NUMBER
                ps number on all node. [default: 0]
        --worker_number WOEKER_NUMBER
                worker number on all node. [default: 1]
        --letrain_args Letrain_Args  args for letrain

    """
    env = get_env(args['<scheduler>'])
    args['<program>'] = '/opt/letrain/applications/letrain.pyc'
    args['--node-num'] = len(env['nodelist'])
    args['--cpu-num'] = 1
    args['--gpu-num'] = 0
    args['--task-index'] = 'task_index'
    args['--job-name'] = 'job_name'
    args['--ps-args'] = 'ps_hosts'
    args['--ps-job-name'] = 'ps'
    args['--wk-args'] = 'worker_hosts'
    args['--wk-job-name'] = 'worker'
    if args['--node-num'] == 1:
        args['--ps_number'] = 0
        args['--worker_number'] = 1
    new_args = check_args(args, env)
    for k in new_args:
        args[k] = new_args[k]
    worker_num = int(args['--worker_number'])
    node_num = args['--node-num']
    worker_num_per_node = int(worker_num / node_num)
    gpu_num = int(args['--gpu-num'] / worker_num_per_node)
    if gpu_num:
        num_clones = " --num_clones=" + str(gpu_num)
    else:
        num_clones = ""
    args['<args>'] = args['--letrain_args'] + num_clones
    base_tensorflow(args, env)


if __name__ == '__main__':
    run_deep_learning()
