# Copyright 2022-present Lenovo
# Confidential and Proprietary

"""
Examples:
    python dataset_analysis.py --workspace="/home/xxx" \
                               --scenario="image_classification" \
                               --location="/home/xxx" \
                               --dataset_format=default
"""

import argparse
import json
import os.path
import importlib
from abc import abstractmethod


class DatasetAnalysisBase(object):

    def __init__(self, ds_format):
        self.ds_format = ds_format

    def check_structure(self, location, dirlist, filelist):
        for dir_name in dirlist:
            dir_path = os.path.join(location, dir_name)
            if not os.path.isdir(dir_path):
                print("[Error]: Dataset format is invalid, "
                      "lacking of dir {}".format(dir_path))
                exit(1)
        for file_name in filelist:
            file_path = os.path.join(location, file_name)
            if not os.path.isfile(file_path):
                print("[Error]: Dataset format is invalid, "
                      "lacking of file {}".format(file_path))
                exit(1)
        return

    def check_empty_dataset(self, dataset_path, sample_num):
        if sample_num == 0:
            print(f"[ERROR]: Cannot find any samples in {dataset_path}")
            exit(1)

    def get_sample_num(self, base_path, allowed_suffix):
        sample_num = 0
        for root, dirs, files in os.walk(base_path):
            for file in files:
                if os.path.splitext(file)[-1] in allowed_suffix:
                    sample_num += 1
        return sample_num

    def get_dir_size_b(self, path):
        get_size = os.path.getsize
        if os.path.isdir(path):
            return sum([sum(
                [get_size(os.path.join(root, name))
                 for name in files])
                for root, dirs, files in os.walk(path)])
        else:
            return 0

    def get_file_type(self, data_dir):
        def _get_file_type(data_dir):
            _files = []
            type_list = ['jpg', 'png', 'jpeg']
            file_list = os.listdir(data_dir)
            for i in range(0, len(file_list)):
                path = os.path.join(data_dir, file_list[i])
                if os.path.isdir(path):
                    _files.extend(_get_file_type(path))
                if os.path.isfile(path):
                    ftype = (path.split('/')[-1]).split('.')[-1]
                    if ftype in type_list:
                        raise Exception(ftype)
            return _files

        try:
            ftype = ''
            _get_file_type(data_dir)
        except Exception as e:
            ftype = e
        ftype = '*.' + str(ftype)
        return ftype

    @abstractmethod
    def verify_format(self, dataset_path):
        pass


def parse_args():
    """Parse input arguments."""
    parser = argparse.ArgumentParser(description='Dataset analysis script')
    parser.add_argument('--workspace', help='Dataset analysis result location',
                        required=True, type=str)
    parser.add_argument('--location', dest='location',
                        help='Dataset project path',
                        required=True, type=str)
    parser.add_argument('--scenario', dest='scenario',
                        help='Dataset type',
                        required=True, type=str)
    parser.add_argument('--dataset_format', help='Dataset format',
                        default='default', type=str)

    return parser.parse_args()


def check_args(args):
    location = args.location
    if not os.path.isdir(location):
        print("[Invalid parameter]: The %s path does not exist "
              "or is not a folder" % location)
        exit(1)


if __name__ == '__main__':
    args = parse_args()
    check_args(args)
    scenario = args.scenario

    if scenario not in (
        'image_classification',
        'object_detection',
        'instance_segmentation',
        'medical_image_segmentation',
        'image_gan',
        'text_classification',
        'memory_network',
        'seq2seq'
    ):
        print("[Invalid parameter]: Invalid dataset scenario.")
        exit(1)

    scenario_module = importlib.import_module(scenario)
    analysis = getattr(scenario_module, 'DatasetAnalysis')

    analysis_result = analysis(ds_format=args.dataset_format).verify_format(
        args.location)

    with open(
        os.path.join(args.workspace, "dataset_analysis.txt"),
        'w', encoding='utf-8'
    ) as f:
        json.dump(analysis_result, f)
    print("[Success]: Dataset analysis results: %s" % analysis_result)
