# Copyright 2022-present Lenovo
# Confidential and Proprietary

import os
import json
import re

from dataset_analysis import DatasetAnalysisBase


class DatasetAnalysis(DatasetAnalysisBase):
    def verify_format(self, dataset_path):
        train, test, valid = self.load_custom_data(dataset_path)
        data = train + test + valid

        sample_num = len(data)

        self.check_empty_dataset(dataset_path, sample_num)
        size = self.get_dir_size_b(dataset_path)
        return {
            "size_b": size,
            "label_num": -1,
            "sample_num": sample_num,
            "label_statistics": json.dumps({
                "Size": size,
                "Samples": sample_num
            })
        }

    def load_custom_data(self, dataset_path, only_supporting=False):
        files = os.listdir(dataset_path)
        files = [os.path.join(dataset_path, f) for f in files]
        files = [f for f in files if not os.path.isdir(f)]
        train_data, test_data, valid_data = [], [], []
        test_file = [f for f in files
                     if 'test' in os.path.relpath(f, dataset_path)]
        if test_file:
            test_data = self.get_stories(test_file[0], only_supporting)
        train_file = [f for f in files
                      if 'train' in os.path.relpath(f, dataset_path)]
        if train_file:
            train_data = self.get_stories(train_file[0], only_supporting)
        valid_file = [f for f in files
                      if 'valid' in os.path.relpath(f, dataset_path)]
        if valid_file:
            valid_data = self.get_stories(valid_file[0], only_supporting)
        return train_data, test_data, valid_data

    def tokenize(self, sent):
        """
        Return the tokens of a sentence including punctuation.
        >>> tokenize('Bob dropped the apple. Where is the apple?')
        ['Bob', 'dropped', 'the', 'apple', '.', 'Where', 'is', 'the',
        'apple', '?']
        """
        return [x.strip() for x in
                re.split('(\W+)+', sent) if x.strip()]  # noqa

    def parse_stories(self, lines, only_supporting=False):
        """
        Parse stories provided in the bAbI tasks format
        If only_supporting is true, only the sentences that
         support the answer are kept.
        """
        data = []
        story = []
        for line in lines:
            line = str.lower(line)
            nid, line = line.split(' ', 1)
            nid = int(nid)
            if nid == 1:
                story = []
            if '\t' in line:  # question
                q, a, supporting = line.split('\t')
                q = self.tokenize(q)
                # a = tokenize(a)
                # answer is one vocab word even if it's actually multiple words
                a = [a]
                substory = None

                # remove question marks
                if q[-1] == "?":
                    q = q[:-1]

                if only_supporting:
                    # Only select the related substory
                    supporting = list(map(int, supporting.split()))
                    substory = [story[i - 1] for i in supporting]
                else:
                    # Provide all the substories
                    substory = [x for x in story if x]

                data.append((substory, q, a))
                story.append('')
            else:  # regular sentence
                # remove periods
                sent = self.tokenize(line)
                if sent[-1] == ".":
                    sent = sent[:-1]
                story.append(sent)
        return data

    def get_stories(self, f, only_supporting=False):
        """
        Given a file name, read the file, retrieve the stories,
         and then convert the sentences into a single story.
        If max_length is supplied, any stories longer than max_length
         tokens will be discarded.
        """
        with open(f) as f:
            return self.parse_stories(f.readlines(),
                                      only_supporting=only_supporting)
