# Copyright 2021-present Lenovo
# Confidential and Proprietary

import os
import sys

import numpy as np
import six
import tensorflow.compat.v1 as tf
tf.disable_v2_behavior()

sys.path += ["/opt/letrain/engine/automl/learner/modelset/tfefficientdet/"]
import collections

from visualize.vis_utils import (STANDARD_COLORS,
                                 _get_multiplier_for_color_randomness,
                                 draw_bounding_box_on_image_array,
                                 draw_keypoints_on_image_array,
                                 draw_mask_on_image_array)


def restore_ckpt(sess, ckpt_path, enable_ema=True, export_ckpt=None):
    """Restore variables from a given checkpoint.

    Args:
    sess: a tf session for restoring or exporting models.
    ckpt_path: the path of the checkpoint. Can be a file path or a folder path.
    enable_ema: whether reload ema values or not.
    export_ckpt: whether to export the restored model.
    """
    sess.run(tf.global_variables_initializer())
    if os.path.isdir(ckpt_path):
        ckpt_path = tf.train.latest_checkpoint(ckpt_path)
    if enable_ema:
        ema = tf.train.ExponentialMovingAverage(decay=0.0)
        ema_vars = get_ema_vars()
        var_dict = ema.variables_to_restore(ema_vars)
        ema_assign_op = ema.apply(ema_vars)
    else:
        var_dict = get_ema_vars()
        ema_assign_op = None
    tf.train.get_or_create_global_step()
    sess.run(tf.global_variables_initializer())
    saver = tf.train.Saver(var_dict, max_to_keep=1)
    saver.restore(sess, ckpt_path)

    if export_ckpt:
        print('export model to {}'.format(export_ckpt))
        if ema_assign_op is not None:
            sess.run(ema_assign_op)
        saver = tf.train.Saver(max_to_keep=1, save_relative_paths=True)
        saver.save(sess, export_ckpt)


def get_ema_vars():
    """Get all exponential moving average (ema) variables."""
    ema_vars = tf.trainable_variables() + tf.get_collection('moving_vars')
    for v in tf.global_variables():
        # We maintain mva for batch norm moving mean and variance as well.
        if 'moving_mean' in v.name or 'moving_variance' in v.name:
            ema_vars.append(v)
    return list(set(ema_vars))


def visualize_image(image,
                    boxes,
                    classes,
                    scores,
                    id_mapping,
                    min_score_thresh=0.2,
                    max_boxes_to_draw=50,
                    line_thickness=4,
                    **kwargs):
    """Visualizes a given image.

    Args:
        image: a image with shape [H, W, C].
        boxes: a box prediction with shape [N, 4]
            ordered [ymin, xmin, ymax, xmax].
        classes: a class prediction with shape [N].
        scores: A list of float value with shape [N].
        id_mapping: a dictionary from class id to name.
            min_score_thresh: minimal score for showing.
            If claass probability is below this threshold,
            then the object will not show up.
        max_boxes_to_draw: maximum bounding box to draw.
        line_thickness: how thick is the bounding box line.
        **kwargs: extra parameters.

    Returns:
        output_image: an output image with annotated boxes and classes.
    """
    category_index = {k: {'id': k, 'name': id_mapping[k]} for k in id_mapping}
    img = np.array(image)
    _, detection_res = visualize_boxes_and_labels_on_image_array(
        img,
        boxes,
        classes,
        scores,
        category_index,
        min_score_thresh=min_score_thresh,
        max_boxes_to_draw=max_boxes_to_draw,
        line_thickness=line_thickness,
        **kwargs)
    return img, detection_res


def visualize_boxes_and_labels_on_image_array(
        image,
        boxes,
        classes,
        scores,
        category_index,
        instance_masks=None,
        instance_boundaries=None,
        keypoints=None,
        keypoint_edges=None,
        track_ids=None,
        use_normalized_coordinates=False,
        max_boxes_to_draw=20,
        min_score_thresh=.5,
        agnostic_mode=False,
        line_thickness=4,
        groundtruth_box_visualization_color='black',
        skip_boxes=False,
        skip_scores=False,
        skip_labels=False,
        skip_track_ids=False):
    """Overlay labeled boxes on an image with formatted scores and label names.

    This function groups boxes that correspond to the same location
    and creates a display string for each detection and overlays these
    on the image. Note that this function modifies the image in place,
    and returns that same image.

    Args:
        image: uint8 numpy array with shape (img_height, img_width, 3)
        boxes: a numpy array of shape [N, 4]
        classes: a numpy array of shape [N].
            Note that class indices are 1-based,
            and match the keys in the label map.
        scores: a numpy array of shape [N] or None.  If scores=None, then
        this function assumes that the boxes to be plotted are groundtruth
        boxes and plot all boxes as black with no classes or scores.
        category_index: a dict containing category dictionaries (each holding
            category index `id` and category name `name`)
            keyed by category indices.
        instance_masks: a numpy array of shape [N, image_height, image_width]
            with values ranging between 0 and 1, can be None.
        instance_boundaries: a numpy array of shape
            [N, image_height, image_width] with values ranging between 0 and 1,
            can be None.
        keypoints: a numpy array of shape [N, num_keypoints, 2], can
            be None
        keypoint_edges: A list of tuples with keypoint indices that specify
            which keypoints should be connected by an edge,
            e.g. [(0, 1), (2, 4)] draws edges from keypoint 0 to 1
            and from keypoint 2 to 4.
        track_ids: a numpy array of shape [N] with unique track ids.
            If provided, color-coding of boxes will be determined by
            these ids, and not the class indices.
        use_normalized_coordinates: whether boxes is to be interpreted as
        normalized coordinates or not.
        max_boxes_to_draw: maximum number of boxes to visualize.  If None, draw
        all boxes.
        min_score_thresh: minimum score threshold for a box to be visualized
        agnostic_mode: boolean (default: False) controlling whether to evaluate
            in class-agnostic mode or not.  This mode will display scores
            but ignore classes.
        line_thickness: integer (default: 4) controlling line width of
            the boxes.groundtruth_box_visualization_color: box color for
            visualizing groundtruth boxes
        skip_boxes: whether to skip the drawing of bounding boxes.
        skip_scores: whether to skip score when drawing a single detection
        skip_labels: whether to skip label when drawing a single detection
        skip_track_ids: whether to skip track id when drawing
            a single detection.

    Returns:
        uint8 numpy array with shape (img_height, img_width, 3)
        with overlaid boxes.
    """
    # Create a display string (and color) for every box location,
    # group any boxes that correspond to the same location.
    box_to_display_str_map = collections.defaultdict(list)
    box_to_color_map = collections.defaultdict(str)
    box_to_instance_masks_map = {}
    box_to_instance_boundaries_map = {}
    box_to_keypoints_map = collections.defaultdict(list)
    box_to_track_ids_map = {}
    detection_res = []
    if not max_boxes_to_draw:
        max_boxes_to_draw = boxes.shape[0]
    for i in range(boxes.shape[0]):
        res = {}
        if max_boxes_to_draw == len(box_to_color_map):
            break
        if scores is None or scores[i] > min_score_thresh:
            box = tuple(boxes[i].tolist())
            box_to_instance_masks_map = _box_to_dict(box_to_instance_masks_map,
                                                     box, i, instance_masks)
            box_to_instance_boundaries_map = _box_to_dict(
                box_to_instance_boundaries_map, box, i, instance_boundaries)
            box_to_keypoints_map = _box_to_dict(box_to_keypoints_map, box, i,
                                                keypoints, True)
            box_to_track_ids_map = _box_to_dict(box_to_track_ids_map, box, i,
                                                track_ids)
            if scores is None:
                box_to_color_map[box] = groundtruth_box_visualization_color
                res = get_res(box, scores[i], category_index, classes[i])
            else:
                box_to_display_str_map[box].append(
                    get_display_str(skip_labels, agnostic_mode, classes, i,
                                    category_index, scores, skip_scores,
                                    track_ids))
                res, box_to_color_map = _box_to_color_map(
                    res, box_to_color_map, agnostic_mode, box, scores,
                    category_index, classes, track_ids, i)
        if res:
            detection_res.append(res)
    _draw_all_box(box_to_color_map, instance_masks, image,
                  box_to_instance_masks_map, instance_boundaries,
                  box_to_instance_boundaries_map, skip_boxes, line_thickness,
                  box_to_display_str_map, use_normalized_coordinates,
                  keypoints, box_to_keypoints_map, keypoint_edges)
    return image, detection_res


def get_res(box, score, category_index, class_id):
    ymin, xmin, ymax, xmax = box
    return {
        "left": int(round(xmin)),
        "top": int(round(ymin)),
        "width": int(round(xmax)) - int(round(xmin)),
        "height": int(round(ymax)) - int(round(ymin)),
        "probability": round(float(score), 4),
        "label": category_index[class_id]['name']
    }


def _box_to_dict(map, box, idx, args, is_extend=False):
    if args is not None:
        if is_extend:
            map[box].extend(args[idx])
        else:
            map[box] = args[idx]
    return map


def get_display_str(skip_labels, agnostic_mode, classes, idx, category_index,
                    scores, skip_scores, track_ids):
    display_str = ''
    if not skip_labels:
        if not agnostic_mode:
            if classes[idx] in six.viewkeys(category_index):
                class_name = category_index[classes[idx]]['name']
            else:
                class_name = 'N/A'
            display_str = str(class_name)
    if not skip_scores:
        if not display_str:
            display_str = '{} %'.format(int(100 * scores[idx]))
        else:
            display_str = '{}: {} %'.format(display_str,
                                            int(100 * scores[idx]))
    if not skip_scores and track_ids is not None:
        if not display_str:
            display_str = 'ID {}'.format(track_ids[idx])
        else:
            display_str = '{}: ID {}'.format(display_str, track_ids[idx])
    return display_str


def _box_to_color_map(res, map, agnostic_mode, box, scores, category_index,
                      classes, track_ids, idx):
    if agnostic_mode:
        map[box] = 'DarkOrange'
        res = get_res(box, scores[idx], category_index, classes[idx])
    elif track_ids is not None:
        prime_multipler = _get_multiplier_for_color_randomness()
        map[box] = STANDARD_COLORS[(prime_multipler * track_ids[idx]) %
                                   len(STANDARD_COLORS)]
        res = get_res(box, scores[idx], category_index, classes[idx])
    else:
        map[box] = STANDARD_COLORS[classes[idx] % len(STANDARD_COLORS)]
        res = get_res(box, scores[idx], category_index, classes[idx])
    return res, map


def _draw_all_box(box_to_color_map, instance_masks, image,
                  box_to_instance_masks_map, instance_boundaries,
                  box_to_instance_boundaries_map, skip_boxes, line_thickness,
                  box_to_display_str_map, use_normalized_coordinates,
                  keypoints, box_to_keypoints_map, keypoint_edges):
    # Draw all boxes onto image.
    for box, color in box_to_color_map.items():
        ymin, xmin, ymax, xmax = box
        if instance_masks is not None:
            draw_mask_on_image_array(image,
                                     box_to_instance_masks_map[box],
                                     color=color)
        if instance_boundaries is not None:
            draw_mask_on_image_array(image,
                                     box_to_instance_boundaries_map[box],
                                     color='red',
                                     alpha=1.0)
        draw_bounding_box_on_image_array(
            image,
            ymin,
            xmin,
            ymax,
            xmax,
            color=color,
            thickness=0 if skip_boxes else line_thickness,
            display_str_list=box_to_display_str_map[box],
            use_normalized_coordinates=use_normalized_coordinates)
        if keypoints is not None:
            draw_keypoints_on_image_array(
                image,
                box_to_keypoints_map[box],
                color=color,
                radius=line_thickness / 2,
                use_normalized_coordinates=use_normalized_coordinates,
                keypoint_edges=keypoint_edges,
                keypoint_edge_color=color,
                keypoint_edge_width=line_thickness // 2)
