# Copyright 2021-present Lenovo
# Confidential and Proprietary

import json
import os
import shutil
import sys
import zipfile
from uuid import uuid4
from datetime import datetime
from subprocess import PIPE, run

import numpy as np
import pycuda.autoinit  # noqa: F401
# This import causes pycuda to automatically manage CUDA context
# creation and cleanup.
import pycuda.driver as cuda
import tensorflow.compat.v1 as tf
tf.disable_v2_behavior()
import tensorrt as trt
from numpy.lib.npyio import load  # noqa: F401
from tensorflow.python.tools import freeze_graph
from trt_common import load_image, load_labels

# You can set the logger severity higher to suppress messages
# (or lower to display more messages).

parentpath = '/opt/letrain/'
path_list = [
    parentpath,
    parentpath + 'lib',
    parentpath + 'engine',
    parentpath + 'inference',
    parentpath + 'applications',
    os.path.join(parentpath, "applications/frcnn")
]
sys.path += path_list

from nets import nets_factory

tf.app.flags.DEFINE_string('model_path', '', '')
tf.app.flags.DEFINE_string('model_name', '', '')
tf.app.flags.DEFINE_string('output_path', '.',
                           'output dir for plan file.')
tf.app.flags.DEFINE_string('trt_mode', 'float32', '')
tf.app.flags.DEFINE_string('calib_ds', '',
                           'calibration dataset used for trt_mode int8.')
tf.app.flags.DEFINE_integer('engine_batch_size', 8,
                            'tensorRT engine max batch size for inference.')
tf.app.flags.DEFINE_integer('calibration_batch_size', 8,
                            'batch size for calibrate int8 engine.')
tf.app.flags.DEFINE_bool('use_server', False, 'whether to add server files.')


FLAGS = tf.app.flags.FLAGS
TRT_LOGGER = trt.Logger(trt.Logger.WARNING)


def GiB(val):
    return val * 1 << 28


class Planfile(object):
    def __init__(
        self,
        model_path,
        model_name,
        output_path,
        engine_batch_size=8,
        trt_mode='float32',
        calibration_batch_size=8,
        calibration_dataset_path='',
        use_server=False
    ):
        self.input_node_name = "lico_input_node"
        self.output_node_name = "lico_output_node"
        self.frozen_model_name = "frozen_model"

        self.model_path = model_path
        self.model_name = model_name

        self.output_path = output_path

        self.engine_batch_size = engine_batch_size
        self.trt_mode = trt_mode
        self.calibration_batch_size = calibration_batch_size
        self.calibration_dataset_path = calibration_dataset_path

        self.use_server = use_server

    def _pre_check(self):
        if not os.path.isdir(self.model_path):
            print("[CONVERT_TRT_ERROR]: model path must be directory.")
            exit(1)

        if self.trt_mode.lower() not in ["float32", "int8"]:
            print(f"[CONVERT_TRT_ERROR]: do not support trt mode: "
                  f"{self.trt_mode}.")
            exit(1)

        if self.trt_mode.lower() == "int8" and \
                not self.calibration_dataset_path:
            print("[CONVERT_TRT_ERROR]: trt mode INT8 requires "
                  "calibration_dataset_path.")
            exit(1)

    def _prepare(self):
        self.labels_dict = load_labels(
            os.path.join(self.model_path, 'labels.txt')
        )
        num_classes = len(self.labels_dict)
        self.network_fn = nets_factory.get_network_fn(
            self.model_name, num_classes, is_training=False)
        self.image_size = self.network_fn.default_image_size
        self.tmp_dir = os.path.join(self.output_path, f'tmp_{uuid4()}')
        os.makedirs(self.tmp_dir)
        os.makedirs(os.path.join(self.tmp_dir, 'service'))

    def convert_ckpt_to_onnx(self):
        print("[INFO] start converting checkpoint to onnx", flush=True)
        ckpt_path = tf.train.latest_checkpoint(self.model_path)
        if not ckpt_path:
            print("[CONVERT_CHECKPOINT_ERROR]: cannot find any checkpoint "
                  "files under model path.")
            exit(1)

        tf.reset_default_graph()
        image_place = tf.compat.v1.placeholder(
            tf.float32,
            [self.engine_batch_size, self.image_size, self.image_size, 3],
            name=self.input_node_name
        )
        logits, _ = self.network_fn(image_place)
        tf.identity(logits, name=self.output_node_name)

        frozen_pb = os.path.join(
            self.output_path,
            self.frozen_model_name + '.pb'
        )
        try:
            with tf.compat.v1.Session() as sess:
                # Save the graph and generate model.pb file in the output path
                # model.pb file will be given as input_graph to the
                # freeze_graph function
                tf.io.write_graph(sess.graph_def, self.output_path, 'model.pb')

                freeze_graph.freeze_graph(
                    input_graph=os.path.join(self.output_path, 'model.pb'),
                    input_saver='',
                    input_binary=False,
                    input_checkpoint=ckpt_path,
                    output_node_names=self.output_node_name,
                    restore_op_name='save/restore_all',
                    filename_tensor_name='save/Const:0',
                    output_graph=frozen_pb,
                    clear_devices=True,
                    initializer_nodes=''
                )
        except Exception as e:
            print("[FREEZE_PB_ERROR]: {0}.".format(e))
            exit(1)
        onnx_file_path = os.path.join(
            self.output_path, self.frozen_model_name + '.onnx')

        process = run(
            ["python", "-m", "tf2onnx.convert", "--input", f"{frozen_pb}",
             "--inputs", f"{self.input_node_name}:0",
             "--outputs", f"{self.output_node_name}:0",
             "--output", f"{onnx_file_path}"],
            stdout=PIPE,
            stderr=PIPE,
            timeout=120
        )
        if process.returncode != 0:
            print("[CONVERT_TO_UFF_ERROR]: {0}".format(str(process.stderr)))
            exit(1)
        return onnx_file_path

    def build_engine_by_onnx(self, onnx_file_path, calib=None):
        explicit_batch = 1 << int(
            trt.NetworkDefinitionCreationFlag.EXPLICIT_BATCH)

        with trt.Builder(TRT_LOGGER) as builder, \
                builder.create_network(explicit_batch) as network, \
                builder.create_builder_config() as config, \
                trt.OnnxParser(network, TRT_LOGGER) as parser:
            # builder.max_workspace_size = GiB(1)
            config.max_workspace_size = GiB(1)
            builder.max_batch_size = self.engine_batch_size

            if calib:
                config.set_flag(trt.BuilderFlag.INT8)
                config.int8_calibrator = calib

            # Parse onnx model from file
            parser.parse_from_file(onnx_file_path)
            # Build and return an engine.
            engine = builder.build_engine(network, config)
            print("[INFO] build engine finished.")
        return engine

    @staticmethod
    def save_engine(plan_file, engine):
        with open(plan_file, "wb") as f:
            f.write(engine.serialize())
        return plan_file

    def packaging_files(self):
        try:
            print("[INFO] start packaging export files.", flush=True)
            # 1. copy sample code to tmp_dir
            shutil.copy(
                '/opt/letrain-run-script/export/sample.py',
                os.path.join(self.tmp_dir, 'sample.py')
            )
            # 2. copy label.txt to tmp_dir
            shutil.copy(
                os.path.join(self.model_path, 'labels.txt'),
                self.tmp_dir
            )
            # 3. clean up useless files: such as some pb, uff files.
            for file_name in ['model.pb',
                              'calibration.cache',
                              self.frozen_model_name + '.pb',
                              self.frozen_model_name + '.onnx']:
                this_file = os.path.join(self.output_path, file_name)
                if os.path.exists(this_file):
                    os.remove(this_file)

            # 4. add server files to service dir
            if self.use_server:
                shutil.copy(
                    '/opt/letrain-run-script/export/trt_service.py',
                    os.path.join(self.tmp_dir, 'service', 'trt_service.py')
                )
                shutil.copy(
                    '/opt/letrain-run-script/export/trt_service_runner.py',
                    os.path.join(self.tmp_dir, 'service',
                                 'trt_service_runner.py')
                )
                shutil.copy(
                    '/opt/letrain-run-script/deploy_service/trt_infe_serv.py',
                    os.path.join(self.tmp_dir, 'service', 'trt_inference.py')
                )
            # 5. write model_name in file
            with open(os.path.join(self.tmp_dir, 'model_define.json'),
                      'w', encoding='UTF-8') as f:
                model_define = {
                    'model_name': self.model_name,
                    'category': 'TensorRT'
                }
                f.write(json.dumps(model_define))

            # 6. make zip file
            zip_name = f'model_TRT_{self.trt_mode}_' \
                       f'{int(datetime.now().timestamp())}.zip'
            with zipfile.ZipFile(
                    os.path.join(self.output_path, zip_name),
                    'w',
                    zipfile.ZIP_DEFLATED
            ) as zf:
                for path, dirnames, filenames in os.walk(self.tmp_dir):
                    for filename in filenames:
                        zf.write(
                            os.path.join(path, filename),
                            os.path.join(
                                path.replace(self.tmp_dir, ''),
                                filename
                            )
                        )
            print(f"[INFO] Finished. You can download files from "
                  f"{os.path.join(self.output_path, zip_name)}")
        finally:
            shutil.rmtree(self.tmp_dir)

    def create_planfile(self):
        self._pre_check()
        self._prepare()

        onnx_file_path = self.convert_ckpt_to_onnx()
        print(f"[INFO] generate planfile with mode {self.trt_mode}")
        if self.trt_mode.lower() == "float32":
            engine = self.build_engine_by_onnx(onnx_file_path)
        else:  # trt_mode == "int8":
            calib_file_sample = []
            for filepath, dirnames, filenames in os.walk(
                    self.calibration_dataset_path):
                for filename in filenames:
                    calib_file_sample.append(load_image(
                        os.path.join(filepath, filename),
                        self.image_size
                    ))
            calib = EntropyCalibrator(
                calib_file_sample,
                os.path.join(self.output_path, "calibration.cache"),
                self.calibration_batch_size,
            )
            engine = self.build_engine_by_onnx(onnx_file_path, calib)

        self.save_engine(
            os.path.join(self.tmp_dir, self.model_name + ".engine"),
            engine
        )
        self.packaging_files()


class EntropyCalibrator(trt.IInt8EntropyCalibrator2):
    def __init__(self, training_data, cache_file, batch_size=1):
        trt.IInt8EntropyCalibrator2.__init__(self)

        self.cache_file = cache_file

        self.data = training_data
        self.batch_size = batch_size
        self.current_index = 0

        # Allocate enough memory for a whole batch.
        self.device_input = cuda.mem_alloc(
            self.data[0].nbytes * self.batch_size)

    def get_batch_size(self):
        return self.batch_size

    def get_batch(self, names):
        if self.current_index + self.batch_size > len(self.data):
            return None

        current_batch = int(self.current_index / self.batch_size)
        if current_batch % 10 == 0:
            print("Calibrating batch {:}, containing {:} images".format(
                current_batch, self.batch_size))

        batch = np.concatenate(
            self.data[self.current_index:self.current_index + self.batch_size])
        cuda.memcpy_htod(self.device_input, batch)
        self.current_index += self.batch_size
        return [self.device_input]

    def read_calibration_cache(self):
        if os.path.exists(self.cache_file):
            with open(self.cache_file, "rb") as f:
                return f.read()

    def write_calibration_cache(self, cache):
        with open(self.cache_file, "wb") as f:
            f.write(cache)


if __name__ == "__main__":
    print(f"[INFO] Params: \n"
          f"model_path: {FLAGS.model_path} \n"
          f"model_name: {FLAGS.model_name} \n"
          f"output_path: {FLAGS.output_path} \n"
          f"trt_mode: {FLAGS.trt_mode} \n"
          f"calibration_dataset_path: {FLAGS.calib_ds} \n"
          f"engine_batch_size: {FLAGS.engine_batch_size} \n"
          f"calibration_batch_size: {FLAGS.calibration_batch_size} \n"
          f"use_server: {FLAGS.use_server}")
    pf = Planfile(
        model_path=FLAGS.model_path,
        model_name=FLAGS.model_name,
        output_path=FLAGS.output_path,
        trt_mode=FLAGS.trt_mode,
        calibration_dataset_path=FLAGS.calib_ds,
        engine_batch_size=FLAGS.engine_batch_size,
        calibration_batch_size=FLAGS.calibration_batch_size,
        use_server=FLAGS.use_server
    )
    pf.create_planfile()
