# Copyright 2022-present Lenovo
# Confidential and Proprietary
import json
import os
import sys
from shutil import copyfile

import keras.backend as K
import tensorflow as tf
from export_model import BaseExportService
from tensorflow.python.platform import gfile
from tensorflow.python.saved_model import tag_constants
from tensorflow.python.saved_model.signature_def_utils_impl import \
    predict_signature_def

parentpath = '/opt/letrain/'
path_list = [parentpath,
             parentpath + 'lib',
             parentpath + 'engine',
             parentpath + 'inference',
             parentpath + 'applications',
             parentpath + 'applications/maskrcnn',
             parentpath + 'applications/maskrcnn/samples',
             parentpath + 'applications/maskrcnn/samples/coco']
sys.path += path_list

import mrcnn.model as modellib
# Import COCO config
from letrain_keras.config import Config

tf.compat.v1.flags.DEFINE_integer("batch_size", 64, "the number of batch size")
FLAGS = tf.compat.v1.flags.FLAGS

LABELS_FILE = os.path.join(FLAGS.lico_checkpoint_dir, 'instances_train.json')
VERSIONCONTROLL = FLAGS.lico_version_control


class CocoConfig(Config):
    """Configuration for training on MS COCO.
    Derives from the base Config class and overrides values specific
    to the COCO dataset.
    """
    # Give the configuration a recognizable name
    NAME = "coco"

    # We use a GPU with 12GB memory, which can fit two images.
    # Adjust down if you use a smaller GPU.
    IMAGES_PER_GPU = 1

    # Uncomment to train on 8 GPUs (default is 1)
    # GPU_COUNT = 8
    def __init__(self, num_classes, GPU_COUNT=1):
        """Set values of computed attributes."""
        # Number of classes (including background)
        # NUM_CLASSES = 1 + 80  # COCO has 80 classes
        self.NUM_CLASSES = num_classes
        super(CocoConfig, self).__init__()


class InferenceConfig(CocoConfig):
    # Set batch size to 1 since we'll be running inference on
    # one image at a time. Batch size = GPU_COUNT * IMAGES_PER_GPU
    GPU_COUNT = 1
    IMAGES_PER_GPU = 1


class ExportService(BaseExportService):
    def __init__(self):
        self.pre_check()
        if not gfile.Exists(LABELS_FILE):
            raise ValueError("label file: %s is not found." % LABELS_FILE)

    def find_last(self, model_dir):
        """ Find the last chkpnt file of the last trained model.
            Returns: The path of the last checkpoint file
        """
        # Get directory names. Each directory corresponds to a model
        task_names = \
            [t for t in os.listdir(model_dir) if t.startswith('task_0')]
        dir_name = os.path.join(model_dir, task_names[-1])
        # Find the last checkpoint
        checkpoints = next(os.walk(dir_name))[2]
        checkpoints = filter(lambda f: f.startswith("mask_rcnn"), checkpoints)
        checkpoints = sorted(checkpoints)
        if not checkpoints:
            import errno
            raise IOError(
                errno.ENOENT,
                "Could not find weight files in {}".format(dir_name)
            )
        checkpoint = os.path.join(dir_name, checkpoints[-1])
        return checkpoint

    def save_pb(self, export_path_base, model):
        export_path = self.get_export_path(
            export_path_base,
            VERSIONCONTROLL,
            FLAGS.model_version
        )

        builder = tf.saved_model.builder.SavedModelBuilder(export_path)

        signature = predict_signature_def(
            inputs={'images': model.keras_model.input[0],
                    'meta': model.keras_model.input[1],
                    'anchor': model.keras_model.input[2]},
            outputs={'detection': model.keras_model.outputs[0],
                     'mask': model.keras_model.outputs[3]}
        )
        with K.get_session() as sess:
            builder.add_meta_graph_and_variables(
                sess=sess,
                tags=[tag_constants.SERVING],
                signature_def_map={'predict': signature},
                clear_devices=True
            )
            builder.save()

        DESTINATION = export_path.decode() + '/category.json'
        copyfile(LABELS_FILE, DESTINATION)
        self.package_model_into_zip_file(export_path_base, export_path)

    def checkpoint2pb(self):
        coco_annotation_file = LABELS_FILE
        if os.path.exists(coco_annotation_file) and \
                os.path.isfile(coco_annotation_file):
            class_names = ['background']
            dataset = json.load(open(coco_annotation_file, 'r'))
            assert type(dataset) == dict, 'annotation file format {} not ' \
                                          'supported'.format(type(dataset))
            categories = dataset['categories']
            for cg in categories:
                class_names.append(cg['name'])
        else:
            class_names = [
                'BG', 'person', 'bicycle', 'car', 'motorcycle', 'airplane',
                'bus', 'train', 'truck', 'boat', 'traffic light',
                'fire hydrant', 'stop sign', 'parking meter', 'bench', 'bird',
                'cat', 'dog', 'horse', 'sheep', 'cow', 'elephant', 'bear',
                'zebra', 'giraffe', 'backpack', 'umbrella', 'handbag', 'tie',
                'suitcase', 'frisbee', 'skis', 'snowboard', 'sports ball',
                'kite', 'baseball bat', 'baseball glove', 'skateboard',
                'surfboard', 'tennis racket', 'bottle', 'wine glass', 'cup',
                'fork', 'knife', 'spoon', 'bowl', 'banana', 'apple',
                'sandwich', 'orange', 'broccoli', 'carrot', 'hot dog', 'pizza',
                'donut', 'cake', 'chair', 'couch', 'potted plant', 'bed',
                'dining table', 'toilet', 'tv', 'laptop', 'mouse', 'remote',
                'keyboard', 'cell phone', 'microwave', 'oven', 'toaster',
                'sink', 'refrigerator', 'book', 'clock', 'vase', 'scissors',
                'teddy bear', 'hair drier', 'toothbrush'
            ]

        config = InferenceConfig(len(class_names), 1)
        config.display()

        # Create model object in inference mode.
        tmp_dir = os.path.join(
            os.path.split(FLAGS.lico_checkpoint_dir)[0],
            "tmp"
        )
        if not os.path.isdir(tmp_dir):
            os.makedirs(tmp_dir)
        model = modellib.MaskRCNN(
            mode="inference",
            model_dir=tmp_dir,
            config=config
        )
        coco_model_path = self.find_last(FLAGS.lico_checkpoint_dir)

        # Load weights trained on MS-COCO
        model.keras_model.load_weights(coco_model_path, by_name=True)
        self.save_pb(FLAGS.lico_pb_path, model)


if __name__ == '__main__':
    seg_obj = ExportService()
    seg_obj.checkpoint2pb()
